import re

import pandas as pd

from melusine.base import MelusineDetector


# --8<-- [start:detector]
# --8<-- [start:detector_init]
class MyVirusDetector(MelusineDetector):
    OUTPUT_RESULT_COLUMN = "virus_result"
    TMP_DETECTION_INPUT_COLUMN = "detection_input"
    TMP_POSITIVE_REGEX_MATCH = "positive_regex_match"
    TMP_NEGATIVE_REGEX_MATCH = "negative_regex_match"

    def __init__(self, body_column: str, header_column: str):
        self.body_column = body_column
        self.header_column = header_column

        super().__init__(
            input_columns=[self.body_column, self.header_column],
            output_columns=[self.OUTPUT_RESULT_COLUMN],
            name="virus",
        )
        # --8<-- [end:detector_init]

    # --8<-- [start:pre_detect]
    def pre_detect(self, row, debug_mode=False):
        effective_text = row[self.header_column] + "\n" + row[self.body_column]
        row[self.TMP_DETECTION_INPUT_COLUMN] = effective_text

        if debug_mode:
            row[self.debug_dict_col] = {"detection_input": row[self.TMP_DETECTION_INPUT_COLUMN]}

        return row
        # --8<-- [end:pre_detect]

    # --8<-- [start:detect]
    def detect(self, row, debug_mode=False):
        text = row[self.TMP_DETECTION_INPUT_COLUMN]
        positive_regex = r"virus"
        negative_regex = r"corona[ _]virus"

        positive_match = re.search(positive_regex, text)
        negative_match = re.search(negative_regex, text)

        row[self.TMP_POSITIVE_REGEX_MATCH] = bool(positive_match)
        row[self.TMP_NEGATIVE_REGEX_MATCH] = bool(negative_match)

        if debug_mode:
            positive_match_text = (
                positive_match.string[positive_match.start() : positive_match.end()] if positive_match else None
            )
            negative_match_text = (
                positive_match.string[negative_match.start() : negative_match.end()] if negative_match else None
            )
            row[self.debug_dict_col].update(
                {
                    "positive_match_data": {"result": bool(positive_match), "match_text": positive_match_text},
                    "negative_match_data": {"result": bool(negative_match), "match_text": negative_match_text},
                }
            )

        return row
        # --8<-- [end:detect]

    # --8<-- [start:post_detect]
    def post_detect(self, row, debug_mode=False):
        if row[self.TMP_POSITIVE_REGEX_MATCH] and not row[self.TMP_NEGATIVE_REGEX_MATCH]:
            row[self.OUTPUT_RESULT_COLUMN] = True
        else:
            row[self.OUTPUT_RESULT_COLUMN] = False

        return row

    # --8<-- [end:post_detect]


# --8<-- [end:detector]


def run():
    # --8<-- [start:run]
    detector = MyVirusDetector(body_column="body", header_column="header")

    df = pd.DataFrame(
        [
            {"body": "This is a dangerous virus", "header": "test"},
            {"body": "test", "header": "test"},
            {"body": "test", "header": "viruses are dangerous"},
            {"body": "corona virus is annoying", "header": "test"},
        ]
    )
    df.debug = True

    df = detector.transform(df)
    # --8<-- [end:run]

    return df



def test_tutorial002(add_docs_to_pythonpath):
    from docs_src.MelusineDetectors.tutorial001 import run as run001

    _ = run001()


def test_tutorial002(add_docs_to_pythonpath):
    from docs_src.MelusineDetectors.tutorial002 import run as run002

    _ = run002()


def test_tutorial003(add_docs_to_pythonpath):
    from docs_src.MelusineDetectors.tutorial003 import run as run003

    _ = run003()


def test_tutorial004(add_docs_to_pythonpath):
    from docs_src.MelusineDetectors.tutorial004 import run as run004

    _ = run004()



"""
Base classes of the Melusine framework.

Implemented classes: [
    MelusineTransformer,
    MelusineDetector,
    MelusineModel,
    BaseLabelProcessor,
    MissingModelInputFieldError,
    MissingFieldError,
    MelusineFeatureEncoder
]
"""
from __future__ import annotations

import copy
import inspect
import logging
import re
from abc import ABC, abstractmethod
from typing import Any, Callable, Dict, Iterable, TypeVar, Union

import pandas as pd
from sklearn.base import BaseEstimator, TransformerMixin

from melusine.backend import backend
from melusine.io import IoMixin

logger = logging.getLogger(__name__)

# Dataset types supported by Melusine : pandas DataFrame and dicts
MelusineDataset = Union[Dict[str, Any], pd.DataFrame]

# Corresponding items are:
# - Dataset : Pandas DataFrame => Item : Pandas Series
# - Dataset Dict => Item Dict
MelusineItem = Union[Dict[str, Any], pd.Series]
Transformer = TypeVar("Transformer", bound="MelusineTransformer")


class TransformError(Exception):
    """
    Exception raised when an error occurs during the transform operation.
    """


class MelusineTransformer(BaseEstimator, TransformerMixin, IoMixin):
    """
    Define a MelusineTransformer object.

    Is an abstract class.

    It can be a Processor or a Detector.
    """

    def __init__(
        self,
        input_columns: str | Iterable[str],
        output_columns: str | Iterable[str],
        func: Callable | None = None,
    ) -> None:
        """
        Attribute initialization.

        Parameters
        ----------
        input_columns: Union[str, Iterable[str]]
            List of input columns
        output_columns: Union[str, Iterable[str]]
            List of output columns
        func: Callable
            Transform function to be applied
        """
        IoMixin.__init__(self)

        self.input_columns: list[str] = self.parse_column_list(input_columns)
        self.output_columns: list[str] = self.parse_column_list(output_columns)
        self.func = func

    @staticmethod
    def parse_column_list(columns: str | Iterable[str]) -> list[str]:
        """
        Transform a string into a list with a single element.

        Parameters
        ----------
        columns: Union[str, Iterable[str]]
            String or list of strings with column name(s).

        Returns
        -------
        _: list[str]
            A list of column names.
        """
        # Change string into list of strings if necessary
        # "body" => ["body]
        if isinstance(columns, str):
            columns = [columns]
        return list(columns)

    def transform(self, data: MelusineDataset) -> MelusineDataset:
        """
        Transform input data.

        Parameters
        ----------
        data: MelusineDataset
            Input data.

        Returns
        -------
        _: MelusineDataset
            Transformed data (output).
        """
        logger.debug(f"Running transform for {type(self).__name__}")
        if self.func is None:
            raise AttributeError(f"Attribute func of MelusineTransformer {type(self).__name__} should not be None")
        try:
            return backend.apply_transform(
                data=data, input_columns=self.input_columns, output_columns=self.output_columns, func=self.func
            )

        except Exception as exception:
            func_name = self.func.__name__
            class_name = type(self).__name__
            input_columns = self.input_columns
            raise TransformError(
                f"Error in class: '{class_name}' "
                f"with method '{func_name}' "
                f"input_columns: {input_columns}\n"
                f"{str(exception)}"
            ).with_traceback(exception.__traceback__) from exception


class BaseMelusineDetector(MelusineTransformer, ABC):
    """
    Used to define detectors.

    Template Method str based on the MelusineTransformer class.
    """

    def __init__(
        self,
        name: str,
        input_columns: list[str],
        output_columns: list[str],
    ):
        """
        Attributes initialization.

        Parameters
        ----------
        name: str
            Name of the detector.
        input_columns:
            Detector input columns.
        output_columns:
            Detector output columns.
        """
        #  self.name needs to be set before the super class init
        #  Name is used to build the output_columns
        self.name = name

        super().__init__(
            input_columns=input_columns,
            output_columns=output_columns,
        )

    @property
    def debug_dict_col(self) -> str:
        """
        Standard name for the column containing the debug info.

        Typically, a detector may return the following outputs:
        - output_result_col: bool
          > Ex: thanks_result: True
        - output_value_col: Any
          > Ex: thanks_output: "Remerciement plat"
        - output_score_col: float
          > Ex: thanks_score: 0.95
        - (debug) debug_dict_col: Dict[str, Any]
          > Ex: debug_thanks: {"thanks_text": "Merci"}
        """
        return f"debug_{self.name}"

    @property
    @abstractmethod
    def transform_methods(self) -> list[Callable]:
        """
        Specify the sequence of methods to be called by the transform method.

        Returns
        -------
        _: list[Callable]
            List of  methods to be called by the transform method.
        """

    def transform(self, df: MelusineDataset) -> MelusineDataset:
        """
        Re-definition of super().transform() => specific detector's implementation

        Transform input data.

        Parameters
        ----------
        df: MelusineDataset
            Input data.

        Returns
        -------
        _: MelusineDataset
            Transformed data (output).
        """
        logger.debug(f"Running transform for {type(self).__name__}")

        # Debug mode ON?
        debug_mode: bool = backend.check_debug_flag(df)

        # Validate fields of the input data
        self.validate_input_fields(df)

        # Work on a copy of the DataFrame and limit fields to effective input columns
        # data_ = backend.copy(data, fields=self.input_columns)

        # Work on a copy of the DataFrame and keep all columns
        # (too complex to handle model input columns)
        data_ = backend.copy(df)

        # Get list of new columns created by the detector
        return_cols = copy.deepcopy(self.output_columns)

        # Create debug data dict
        if debug_mode:
            data_ = backend.setup_debug_dict(data_, dict_name=self.debug_dict_col)
            return_cols.append(self.debug_dict_col)

        for method in self.transform_methods:
            logger.debug(f"Running transform for {type(self).__name__} ({method.__name__})")
            first_arg_name: str = list(inspect.signature(method).parameters)[0]

            if first_arg_name == "row":
                # Run row-wise method
                data_ = backend.apply_transform(
                    data=data_, input_columns=None, output_columns=None, func=method, debug_mode=debug_mode
                )
            else:
                data_ = method(data_, debug_mode=debug_mode)

        # Add new fields to the original MelusineDataset
        data = backend.add_fields(left=df, right=data_, fields=return_cols)

        return data

    def validate_input_fields(self, data: MelusineDataset) -> None:
        """
        Make sure that all the required input fields are present.

        Parameters
        ----------
        data: MelusineDataset
            Input data.
        """
        input_fields: list[str] = backend.get_fields(data)
        missing_fields: list[str] = [x for x in self.input_columns if x not in input_fields]
        if missing_fields:
            raise MissingFieldError(f"Fields {missing_fields} are missing from the input data")


class MelusineDetector(BaseMelusineDetector, ABC):
    """
    Defines an interface for detectors.
    All detectors used in a MelusinePipeline should inherit from the MelusineDetector class and
    implement the abstract methods.
    This ensures homogeneous coding style throughout the application.
    Alternatively, melusine user's can define their own Interface (inheriting from the BaseMelusineDetector)
    to suit their needs.
    """

    @property
    def transform_methods(self) -> list[Callable]:
        """
        Specify the sequence of methods to be called by the transform method.

        Returns
        -------
        _: list[Callable]
            List of  methods to be called by the transform method.
        """
        return [self.pre_detect, self.detect, self.post_detect]

    @abstractmethod
    def pre_detect(self, row: MelusineItem, debug_mode: bool = False) -> MelusineItem:
        """What needs to be done before detection."""

    @abstractmethod
    def detect(self, row: MelusineItem, debug_mode: bool = False) -> MelusineItem:
        """Run detection."""

    @abstractmethod
    def post_detect(self, row: MelusineItem, debug_mode: bool = False) -> MelusineItem:
        """What needs to be done after detection (e.g., mapping columns)."""


class MissingFieldError(Exception):
    """
    Exception raised when a missing field is encountered by a MelusineTransformer
    """


class MelusineRegex(ABC):
    """
    Class to standardise text pattern detection using regex.
    """

    REGEX_FLAGS: re.RegexFlag = re.IGNORECASE | re.MULTILINE

    # Match fields
    MATCH_RESULT: str = "match_result"
    NEUTRAL_MATCH_FIELD: str = "neutral_match_data"
    POSITIVE_MATCH_FIELD: str = "positive_match_data"
    NEGATIVE_MATCH_FIELD: str = "negative_match_data"

    # Match data
    MATCH_START: str = "start"
    MATCH_STOP: str = "stop"
    MATCH_TEXT: str = "match_text"

    def __init__(self, substitution_pattern: str = " ", default_match_group: str = "DEFAULT"):
        if not isinstance(substitution_pattern, str) or (len(substitution_pattern) > 1):
            raise ValueError(
                f"Parameter substitution_pattern should be a string of length 1, not {substitution_pattern}"
            )
        self.substitution_pattern = substitution_pattern
        self.default_match_group = default_match_group

    @property
    def regex_name(self) -> str:
        """
        Name of the Melusine regex object.
        Defaults to the class name.
        """
        return getattr(self, "_regex_name", type(self).__name__)

    @property
    @abstractmethod
    def positive(self) -> dict[str, str] | str:
        """
        Define regex patterns required to activate the MelusineRegex.

        Returns:
            _: Regex pattern or dict of regex patterns.
        """

    @property
    def neutral(self) -> dict[str, str] | str | None:
        """
        Define regex patterns to be ignored when running detection.

        Returns:
            _: Regex pattern or dict of regex patterns.
        """
        return None

    @property
    def negative(self) -> dict[str, str] | str | None:
        """
        Define regex patterns prohibited to activate the MelusineRegex.

        Returns:
            _: Regex pattern or dict of regex patterns.
        """
        return None

    @property
    @abstractmethod
    def match_list(self) -> list[str]:
        """
        List of texts that should activate the MelusineRegex.

        Returns:
            _: List of texts.
        """

    @property
    @abstractmethod
    def no_match_list(self) -> list[str]:
        """
        List of texts that should NOT activate the MelusineRegex.

        Returns:
            _: List of texts.
        """

    def _get_match(
        self, text: str, base_regex: str | dict[str, str], regex_group: str | None = None
    ) -> dict[str, list[dict[str, Any]]]:
        """
        Run specified regex on the input text and return a dict with matching group as key.

        Args:
            text: Text to apply regex on.
            base_regex: Regex to apply on text.
            regex_group: Name of the group the regex belongs to.

        Returns:
            Dict of regex matches for each regex group.
        """
        match_data_dict = {}

        if regex_group is None:
            regex_group = self.default_match_group

        if isinstance(base_regex, dict):
            for group, regex in base_regex.items():
                group_match_data = self._get_match(text, regex, group)
                match_data_dict.update(group_match_data)
        else:
            for match in re.finditer(base_regex, text, flags=self.REGEX_FLAGS):
                if not match_data_dict.get(regex_group):
                    match_data_dict[regex_group] = []

                # Get match position
                start, stop = match.span()

                match_data_dict[regex_group].append(
                    {
                        self.MATCH_START: start,
                        self.MATCH_STOP: stop,
                        self.MATCH_TEXT: text[start:stop],
                    }
                )

        return match_data_dict

    def ignore_text(
        self,
        text: str,
        match_data_dict: dict[str, list[dict[str, Any]]],
    ) -> str:
        """
        Replace neutral regex match text with substitution text to ignore it.

        Args:
            text: Input text.
            match_data_dict: Regex match results.

        Returns:
            _: Text with substituions.
        """
        for _, match_list in match_data_dict.items():
            for match_data in match_list:
                start = match_data[self.MATCH_START]
                stop = match_data[self.MATCH_STOP]

                # Mask text to ignore
                text = text[:start] + self.substitution_pattern * (stop - start) + text[stop:]

        return text

    def get_match_result(self, text: str) -> bool:
        """
        Apply MelusineRegex patterns (neutral, negative and positive) on the input text.
        Return a boolean output of the match result.

        Args:
            text: input text.

        Returns:
            _: True if the MelusineRegex matches the input text.
        """
        result = self(text)
        return result[self.MATCH_RESULT]

    def __call__(self, text: str) -> dict[str, Any]:
        """
        Apply MelusineRegex patterns (neutral, negative and positive) on the input text.
        Return a detailed output of the match results as a dict.

        Args:
            text: input text.

        Returns:
            _: Regex match results.
        """
        match_dict = {
            self.MATCH_RESULT: False,
            self.NEUTRAL_MATCH_FIELD: {},
            self.NEGATIVE_MATCH_FIELD: {},
            self.POSITIVE_MATCH_FIELD: {},
        }

        negative_match = False

        if self.neutral:
            neutral_match_data = self._get_match(text=text, base_regex=self.neutral)
            match_dict[self.NEUTRAL_MATCH_FIELD] = neutral_match_data

            text = self.ignore_text(text, neutral_match_data)

        if self.negative:
            negative_match_data = self._get_match(text=text, base_regex=self.negative)
            negative_match = bool(negative_match_data)
            match_dict[self.NEGATIVE_MATCH_FIELD] = negative_match_data

        positive_match_data = self._get_match(text=text, base_regex=self.positive)
        positive_match = bool(positive_match_data)
        match_dict[self.POSITIVE_MATCH_FIELD] = positive_match_data

        match_dict[self.MATCH_RESULT] = positive_match and not negative_match

        return match_dict

    def describe(self, text: str, position: bool = False) -> None:
        """
        User-friendly description of the regex match results.

        Args:
            text: Input text.
            position: If True, print regex match start and stop positions.
        """

        def _describe_match_field(match_field_data: dict[str, list[dict[str, Any]]]) -> None:
            """
            Format and print result description text.

            Args:
                match_field_data: Regex match result for a given field.
            """
            for group, match_list in match_field_data.items():
                for match_dict in match_list:
                    print(f"{indent}({group}) {match_dict[self.MATCH_TEXT]}")
                    if position:
                        print(f"{indent}start: {match_dict[self.MATCH_START]}")
                        print(f"{indent}stop: {match_dict[self.MATCH_STOP]}")

        indent = " " * 4
        match_data = self(text)

        if match_data[self.MATCH_RESULT]:
            print("The MelusineRegex match result is : POSITIVE")
        else:
            print("The MelusineRegex match result is : NEGATIVE")

        if not any(
            [
                match_data[self.NEUTRAL_MATCH_FIELD],
                match_data[self.NEGATIVE_MATCH_FIELD],
                match_data[self.POSITIVE_MATCH_FIELD],
            ]
        ):
            print("The input text did not match anything.")

        if match_data[self.NEUTRAL_MATCH_FIELD]:
            print("The following text was ignored:")
            _describe_match_field(match_data[self.NEUTRAL_MATCH_FIELD])

        if match_data[self.NEGATIVE_MATCH_FIELD]:
            print("The following text matched negatively:")
            _describe_match_field(match_data[self.NEGATIVE_MATCH_FIELD])

        if match_data[self.POSITIVE_MATCH_FIELD]:
            print("The following text matched positively:")
            _describe_match_field(match_data[self.POSITIVE_MATCH_FIELD])

    def test(self) -> None:
        """
        Test the MelusineRegex on the match_list and no_match_list.
        """
        for text in self.match_list:
            match = self(text)
            assert match[self.MATCH_RESULT] is True, f"Expected match for text\n{text}\nObtained: {match}"

        for text in self.no_match_list:
            match = self(text)
            assert match[self.MATCH_RESULT] is False, f"Expected no match for text:\n{text}\nObtained: {match}"

    def __repr__(self) -> str:
        return f"{type(self).__name__}(positive:{self.positive},neutral:{self.neutral},negative:{self.negative})"



from unittest.mock import patch

import pandas as pd
import pytest

transformers = pytest.importorskip("transformers")
from typing import List

from transformers.pipelines.zero_shot_classification import (
    ZeroShotClassificationPipeline,
)


class MockZeroShotClassificationPipeline(ZeroShotClassificationPipeline):
    def __init__(self, task, model, tokenizer):
        pass

    def __call__(self, sequences, candidate_labels, hypothesis_template):
        if isinstance(sequences, List):
            # Standard return with 2 elements
            return [
                {
                    "sequence": sequences[0],
                    "labels": ["négatif", "positif"],
                    "scores": [0.5, 0.5],
                },
                {
                    "sequence": sequences[1],
                    "labels": ["négatif", "positif"],
                    "scores": [0.5, 0.5],
                },
            ]

        if "gentillesse" in sequences:
            return {
                "sequence": sequences,
                "labels": ["positif", "négatif"],
                "scores": [0.9756866097450256, 0.024313366040587425],
            }
        elif "pas satisfait" in sequences:
            return {
                "sequence": sequences,
                "labels": ["négatif", "positif"],
                "scores": [0.7485730648040771, 0.25142696499824524],
            }
        else:
            return {
                "sequence": sequences,
                "labels": ["négatif", "positif"],
                "scores": [0.5, 0.5],
            }


def test_tutorial001(add_docs_to_pythonpath):
    from docs_src.BasicClassification.tutorial001 import run, transformers_standalone

    with patch(
        "docs_src.BasicClassification.tutorial001.pipeline",
        new=MockZeroShotClassificationPipeline,
    ):
        result = transformers_standalone()
        assert isinstance(result, List)

        df = run()
        assert isinstance(df, pd.DataFrame)



"""
Unit test for pipeline.py
"""
import pandas as pd
import pytest

from melusine.base import MelusineTransformer
from melusine.pipeline import MelusinePipeline

# Dummy variables
dum0 = 42
dum1 = "dum"
dum2 = "dumdum"
dum3 = "duuum"


# Define dummy processors
class DummyProcessor(MelusineTransformer):
    def __init__(self, input_columns="a", output_columns=("b",), dummy_attr=dum1):
        super().__init__(input_columns, output_columns, func=self.add_dummy_col)
        self.dummy_attr = dummy_attr

    def add_dummy_col(self, col_a_data):
        return self.dummy_attr


def test_pipeline_with_processors():
    d1 = DummyProcessor()
    d2 = DummyProcessor(output_columns=("c",), dummy_attr=dum2)

    # Create pipeline
    pipe = MelusinePipeline(steps=[("d1", d1), ("d2", d2)], verbose=True)

    # Create data
    df = pd.DataFrame({"a": [dum0, dum0]})

    # Fit the pipeline and transform the data
    df_transformed = pipe.transform(df)

    # Most basic test, check that the pipeline returns a pandas DataFrame
    assert isinstance(df_transformed, pd.DataFrame)
    assert "a" in df_transformed.columns
    assert "b" in df_transformed.columns
    assert "c" in df_transformed.columns

    assert df_transformed["a"].iloc[0] == dum0
    assert df_transformed["b"].iloc[0] == dum1
    assert df_transformed["c"].iloc[0] == dum2


def test_meta_pipeline():
    d1 = DummyProcessor()
    d2 = DummyProcessor(output_columns=("c",), dummy_attr=dum2)
    d3 = DummyProcessor(output_columns=("d",), dummy_attr=dum3)

    # Create pipeline
    pipe1 = MelusinePipeline(steps=[("d1", d1), ("d2", d2)], verbose=True)
    pipe2 = MelusinePipeline(steps=[("d3", d3)])
    meta_pipe = MelusinePipeline(steps=[("pipe1", pipe1), ("pipe2", pipe2)])

    # Create data
    df = pd.DataFrame({"a": [dum0, dum0]})

    # Fit the pipeline and transform the data
    df_transformed = meta_pipe.transform(df)

    # Most basic test, check that the pipeline returns a pandas DataFrame
    assert isinstance(df_transformed, pd.DataFrame)
    assert "a" in df_transformed.columns
    assert "b" in df_transformed.columns
    assert "c" in df_transformed.columns
    assert "d" in df_transformed.columns

    assert df_transformed["a"].iloc[0] == dum0
    assert df_transformed["b"].iloc[0] == dum1
    assert df_transformed["c"].iloc[0] == dum2
    assert df_transformed["d"].iloc[0] == dum3


def test_pipeline_from_config():
    _ = MelusinePipeline.from_config(config_key="my_pipeline")


def test_pipeline_from_config_error():
    with pytest.raises(ValueError, match=r"'config_key' and 'config_dict'"):
        _ = MelusinePipeline.from_config(config_key="x", config_dict={"a": 5})


def test_pipeline_get_config_from_key():
    conf = MelusinePipeline.get_config_from_key(config_key="my_pipeline")
    assert conf


def test_pipeline_from_config_missing_class():
    config_dict = {
        "steps": [
            {
                "class_name": "UnknownClass",
                "module": "melusine.processors",
                "name": "test",
                "parameters": {
                    "mel": "usine",
                },
            },
        ]
    }
    with pytest.raises(AttributeError, match=r"UnknownClass.*melusine.processors"):
        _ = MelusinePipeline.from_config(config_dict=config_dict)


def test_pipeline_input_output_columns():
    d1 = DummyProcessor(input_columns=("a",), output_columns=("b",))
    d2 = DummyProcessor(input_columns=("b",), output_columns=("c",), dummy_attr=dum2)

    # Create pipeline
    pipe = MelusinePipeline(steps=[("d1", d1), ("d2", d2)], verbose=True)

    assert len(pipe.input_columns) == 2
    assert "a" in pipe.input_columns
    assert "b" in pipe.input_columns

    assert len(pipe.output_columns) == 2
    assert "b" in pipe.output_columns
    assert "c" in pipe.output_columns



"""
This module contains classes for the MelusinePipeline object.

Implemented classes: [PipelineConfigurationError, MelusinePipeline]
"""
from __future__ import annotations

import copy
import importlib
from typing import Iterable, TypeVar

from sklearn.pipeline import Pipeline

from melusine import config
from melusine.backend import backend
from melusine.backend.base_backend import Any
from melusine.base import MelusineTransformer
from melusine.io import IoMixin

T = TypeVar("T")


class PipelineConfigurationError(Exception):
    """
    Error raised when an error is found in the pipeline configuration.
    """


class MelusinePipeline(Pipeline):
    """
    This class defines and executes data transformation.

    The MelusinePipeline is built on top of sklearn Pipelines.
    """

    OBJ_NAME: str = "name"
    OBJ_KEY: str = "config_key"
    OBJ_PARAMS: str = "parameters"
    STEPS_KEY: str = "steps"
    OBJ_CLASS: str = "class_name"
    OBJ_MODULE: str = "module"

    def __init__(
        self,
        steps: list[tuple[str, MelusineTransformer]],
        memory: bool | None = None,
        verbose: bool = False,
    ) -> None:
        """
        Initialize attributes.

        Parameters
        ----------
        steps: List[Tuple[str, MelusineTransformer]]
            List of the pipeline steps.
        memory: bool
            If True, cache invariant transformers when running grid searches.
        verbose: bool
            Verbose mode.
        """
        Pipeline.__init__(self, steps=steps, memory=memory, verbose=verbose)

        self.memory = memory
        self.verbose = verbose

    @property
    def input_columns(self) -> list[str]:
        """
        Input fields of the Pipeline.

        Returns
        -------
        _: List[str]
            List of input fields.
        """
        column_set: set[str] = set()
        for _, step in self.steps:
            # UNION between sets
            column_set |= set(step.input_columns)

        return list(column_set)

    @property
    def output_columns(self) -> list[str]:
        """
        Output fields of the Pipeline.

        Returns
        -------
        _: List[str]
            List of output fields.
        """
        column_set: set[str] = set()
        for _, step in self.steps:
            column_set |= set(step.output_columns)

        return list(column_set)

    @classmethod
    def get_obj_class(cls, obj_params: dict[str, Any]) -> Any:
        """
        Get the class object of an instance.

        Parameters
        ----------
        obj_params: Dict[str, Any].

        Returns
        -------
        _: Any
            Class object.
        """
        obj_class_name = obj_params.pop(cls.OBJ_CLASS)
        obj_module = obj_params.pop(cls.OBJ_MODULE)

        obj_class = MelusinePipeline.import_class(obj_class_name, obj_module)

        return obj_class

    @staticmethod
    def import_class(obj_class_name: str, obj_module: str) -> Any:
        """
        Method to import a class dynamically.

        Parameters
        ----------
        obj_class_name: str
            Name of the object to be imported.
        obj_module: str
            Name of the module containing the object to be imported.

        Returns
        -------
        _: Any
            Class object.
        """
        # Import object class from name and module
        module = importlib.import_module(obj_module)
        if not hasattr(module, obj_class_name):
            raise AttributeError(f"Object `{obj_class_name}` cannot be loaded from module `{module}`.")
        obj_class = getattr(module, obj_class_name)
        return obj_class

    @classmethod
    def flatten_pipeline_config(cls, conf: dict[str, Any]) -> dict[str, Any]:
        """
        Flatten nested Melusine Pipelines.

        This makes it easier for the rest of the processing.

        Parameters
        ----------
        conf: Dict[str, Any]
            Base pipeline conf possibly containing nested pipelines.

        Returns
        -------
        _: Dict[str, Any]
            Flattened conf.
        """
        new_conf: list[Any] = list()
        for step in conf[cls.STEPS_KEY]:
            if step.get(cls.OBJ_CLASS, "") == cls.__name__:
                subpipeline_conf = cls.flatten_pipeline_config(step["parameters"])
                new_conf.extend(subpipeline_conf[cls.STEPS_KEY])
            else:
                new_conf.append(step)
        conf[cls.STEPS_KEY] = new_conf

        return conf

    @classmethod
    def from_config(
        cls, config_key: str | None = None, config_dict: dict[str, Any] | None = None, **kwargs: Any
    ) -> MelusinePipeline:
        """
        Instantiate a MelusinePipeline from a config key.

        Parameters
        ----------
        config_key: str
            Key of the pipeline configuration.
        config_dict: dict
            Dict containing the pipeline configuration.

        Returns
        -------
        _: MelusinePipeline
            Pipeline instance.
        """
        init_params = dict()

        # Get config dict
        if config_key and not config_dict:
            raw_config_dict = config[config_key]
            config_dict = cls.parse_pipeline_config(raw_config_dict)

        elif config_dict and not config_key:
            config_dict = cls.parse_pipeline_config(config_dict)
        else:
            raise ValueError("You should specify one and only one of 'config_key' and 'config_dict'")

        # Prepare step list
        steps = list()

        # Load steps meta data
        steps_meta = config_dict.pop(cls.STEPS_KEY)

        # Instantiate transformers
        for obj_meta in steps_meta:
            # Step name
            step_name: str = obj_meta.pop(cls.OBJ_NAME, None)

            # Step class
            obj_class = cls.get_obj_class(obj_meta)

            # Step arguments
            obj_params = obj_meta[cls.OBJ_PARAMS]

            if issubclass(obj_class, IoMixin):
                obj = obj_class.from_config(config_dict=obj_params)
            else:
                raise TypeError(f"Object {obj_class} does not inherit from the SaverMixin class")  # pragma: no cover

            # Add step to pipeline
            steps.append((step_name, obj))

        # Init params
        init_params.update(config_dict)
        init_params.update(kwargs)

        # Instantiate MelusinePipeline object
        return cls(steps=steps, **init_params)

    @classmethod
    def validate_step_config(cls, step: dict[str, Any]) -> dict[str, Any]:
        """
        Validate a pipeline step configuration.

        Parameters
        ----------
        step: Dict with a pipeline step configuration

        Returns
        -------
        _: Validated pipeline step configuration.
        """
        if not step.get(cls.OBJ_CLASS) or not step.get(cls.OBJ_MODULE):
            raise PipelineConfigurationError(
                f"Pipeline step conf should have a {cls.OBJ_MODULE} key and a {cls.OBJ_CLASS} key."
            )

        if step.get(cls.OBJ_KEY):
            return {
                cls.OBJ_CLASS: step[cls.OBJ_CLASS],
                cls.OBJ_MODULE: step[cls.OBJ_MODULE],
                cls.OBJ_KEY: step[cls.OBJ_KEY],
            }

        if not step.get(cls.OBJ_NAME) or not step.get(cls.OBJ_PARAMS):
            raise PipelineConfigurationError(
                f"Pipeline step conf should have a {cls.OBJ_NAME} key and a {cls.OBJ_KEY} key "
                f"(unless a {cls.OBJ_KEY} is specified)."
            )

        if not isinstance(step[cls.OBJ_PARAMS], dict):
            raise PipelineConfigurationError(
                f"The key {cls.OBJ_PARAMS} should be dictionary not {type(step[cls.OBJ_PARAMS])}"
            )

        return {
            cls.OBJ_CLASS: step[cls.OBJ_CLASS],
            cls.OBJ_MODULE: step[cls.OBJ_MODULE],
            cls.OBJ_NAME: step[cls.OBJ_NAME],
            cls.OBJ_PARAMS: step[cls.OBJ_PARAMS],
        }

    @classmethod
    def validate_pipeline_config(cls, pipeline_conf: dict[str, Any]) -> dict[str, Any]:
        """
        Validate a pipeline configuration.

        Parameters
        ----------
        pipeline_conf: Dict with a pipeline configuration

        Returns
        -------
        _: Validated pipeline configuration.
        """
        validated_pipeline_conf: dict[str, Any] = {cls.STEPS_KEY: []}
        steps = pipeline_conf.get(cls.STEPS_KEY)

        if not steps or not isinstance(steps, list):
            raise PipelineConfigurationError(
                f"Pipeline conf should have a {cls.STEPS_KEY} key containing a list of steps."
            )
        else:
            for step in steps:
                validated_pipeline_conf[cls.STEPS_KEY].append(cls.validate_step_config(step))

        return validated_pipeline_conf

    @classmethod
    def parse_pipeline_config(cls, config_dict: dict[str, Any]) -> dict[str, Any]:
        """
        Parse config dict to replace config key by the associated configurations.

        Parameters
        ----------
        config_dict: Dict[str, Any]
            Initial config.

        Returns
        -------
        _: Dict[str, Any]
            Parsed config.
        """
        config_dict = copy.deepcopy(config_dict)

        # Validate raw pipeline conf
        config_dict = cls.validate_pipeline_config(config_dict)

        steps = []
        for step in config_dict[cls.STEPS_KEY]:
            # Step defined from the config
            config_key = step.get(cls.OBJ_KEY)
            if config_key:
                # Use config key as step name
                step[cls.OBJ_NAME] = config_key
                _ = step.pop(cls.OBJ_KEY)
                # Update step parameters
                step[cls.OBJ_PARAMS] = config[config_key]

            # Nested pipeline
            if step[cls.OBJ_CLASS] == cls.__name__:
                raw_nested_pipeline_config = step[cls.OBJ_PARAMS]
                step[cls.OBJ_PARAMS] = cls.parse_pipeline_config(raw_nested_pipeline_config)

            # Add parsed step config to step list
            steps.append(step)

        config_dict[cls.STEPS_KEY] = steps

        return MelusinePipeline.flatten_pipeline_config(config_dict)

    @classmethod
    def get_config_from_key(cls, config_key: str) -> dict[str, Any]:
        """
        Parse config dict to replace config key by the associated configurations.

        Parameters
        ----------
        config_key: Pipeline configuration key

        Returns
        -------
        _: Dict[str, Any]
            Parsed config.
        """
        return cls.parse_pipeline_config(config_dict=config[config_key])

    def validate_input_fields(self, data: Any) -> None:
        """
        Validate input fields prior of executing the pipeline.
        Use the input_columns and output_columns attributes of each step.

        Parameters
        ----------
        data: Any
            Input data.
        """
        active_fields: set[str] = set(backend.get_fields(data))

        for step_name, step in self.steps:
            difference = set(step.input_columns).difference(active_fields)
            if difference:
                raise ValueError(
                    f"Error at step '{step_name}'.\n"
                    f"Fields {difference} should be either:\n"
                    "- Present in input fields or\n"
                    "- Created by a previous pipeline step"
                )

            active_fields |= set(step.output_columns)

    def transform(self, X: Iterable[Any]) -> Iterable[Any]:  # NOSONAR
        """
        Transform input dataset.

        Parameters
        ----------
        X: Dataset
            Input Dataset.

        Returns
        -------
        _: Dataset
            Output Dataset.
        """
        self.validate_input_fields(X)
        return super().transform(X)



"""
Unit tests of the TransferDetector.
"""


import pandas as pd
import pytest
from pandas import DataFrame

from melusine.detectors import TransferDetector
from melusine.message import Message
from melusine.pipeline import MelusinePipeline


def test_instanciation():
    """Instanciation base test."""

    detector = TransferDetector(name="transfer", header_column="det_clean_header", messages_column="messages")
    assert isinstance(detector, TransferDetector)


@pytest.mark.parametrize(
    "row, good_result",
    [
        (
            {
                "reply_text": "tr: Devis habitation",
                "messages": [
                    Message(
                        meta="",
                        text="Bonjour, je vous confirme l'annulation du rdv du 01/01/2022 "
                        + "à 16h. Bien cordialement, John Smith.",
                        tags=[
                            ("HELLO", "Bonjour,"),
                            (
                                "BODY",
                                "je vous confirme l'annulation du rdv du 01/01/2022 " + "à 16h. ",
                            ),
                            ("GREETINGS", "Cordialement, John Smith."),
                        ],
                    )
                ],
            },
            True,
        ),
        (
            {
                "reply_text": "re: Envoi d'un document de la Société Imaginaire",
                "messages": [
                    Message(
                        meta="this is meta",
                        text="Bonjour, je vous confirme l'annulation du rdv du 01/01/2022 "
                        + "à 16h. Bien cordialement, John Smith.",
                        tags=[
                            ("HELLO", "Bonjour,"),
                            (
                                "BODY",
                                "je vous confirme l'annulation du rdv du 01/01/2022 " + "à 16h. ",
                            ),
                            ("GREETINGS", "Cordialement, John Smith."),
                        ],
                    )
                ],
            },
            True,
        ),
        (
            {
                "reply_text": "re: Virement",
                "messages": [
                    Message(
                        meta="",
                        text="Bonjour, je vous confirme l'annulation du rdv du 01/01/2022 "
                        + "à 16h. Bien cordialement, John Smith.",
                        tags=[
                            ("HELLO", "Bonjour,"),
                            (
                                "BODY",
                                "je vous confirme l'annulation du rdv du 01/01/2022 " + "à 16h. ",
                            ),
                            ("GREETINGS", "Cordialement, John Smith."),
                        ],
                    )
                ],
            },
            False,
        ),
        (
            {
                "reply_text": "",
                "messages": [
                    Message(
                        meta="",
                        text="Bonjour, je vous confirme l'annulation du rdv du 01/01/2022 "
                        + "à 16h. Bien cordialement, John Smith.",
                        tags=[
                            ("HELLO", "Bonjour,"),
                            (
                                "BODY",
                                "je vous confirme l'annulation du rdv du 01/01/2022 " + "à 16h. ",
                            ),
                            ("GREETINGS", "Cordialement, John Smith."),
                        ],
                    )
                ],
            },
            False,
        ),
    ],
)
def test_deterministic_detect(row, good_result):
    """Method base test."""

    # Instanciate manually a detector
    detector = TransferDetector(
        name="transfer",
        header_column="det_clean_header",
        messages_column="messages",
    )
    row = detector.detect(row)
    res = row[detector.result_column]
    assert res == good_result


@pytest.mark.parametrize(
    "df_emails, expected_result",
    [
        (
            DataFrame(
                {
                    "det_clean_header": "tr: Rdv",
                    "messages": [
                        [
                            Message(
                                meta="",
                                text="Bonjour, je vous confirme l'annulation du rdv du 01/01/2022 "
                                + "à 16h. Bien cordialement, John Smith.",
                                tags=[
                                    ("HELLO", "Bonjour,"),
                                    (
                                        "BODY",
                                        "je vous confirme l'annulation du rdv du 01/01/2022 " + "à 16h. ",
                                    ),
                                    ("GREETINGS", "Cordialement, John Smith."),
                                ],
                            )
                        ]
                    ],
                }
            ),
            True,
        ),
    ],
)
def test_transform(df_emails, expected_result):
    """Unit test of the transform() method."""

    # Copy for later load/save test
    df_copy = df_emails.copy()

    # Instanciate manually a detector
    detector = TransferDetector(
        name="transfer",
        header_column="det_clean_header",
        messages_column="messages",
    )

    # Get result column name
    res_col = detector.result_column

    # Apply the detector on data
    df_emails = detector.transform(df_emails)

    # Verify result
    result = df_emails[res_col][0]
    assert result == expected_result


@pytest.mark.parametrize(
    "df_emails, expected_result, expected_debug_info",
    [
        (
            DataFrame(
                {
                    "det_clean_header": ["Tr: Suivi de dossier"],
                    "messages": [
                        [
                            Message(
                                meta="",
                                text="Bonjour, je vous confirme l'annulation du rdv du 01/01/2022 "
                                + "à 16h. Bien cordialement, John Smith.",
                                tags=[
                                    ("HELLO", "Bonjour,"),
                                    (
                                        "BODY",
                                        "je vous confirme l'annulation du rdv du 01/01/2022 " + "à 16h. ",
                                    ),
                                    ("GREETINGS", "Cordialement, John Smith."),
                                ],
                            )
                        ]
                    ],
                }
            ),
            True,
            {
                "reply_text": "tr: suivi de dossier",
                "messages[0].meta": "",
                "TransferRegex": {
                    "match_result": True,
                    "negative_match_data": {},
                    "neutral_match_data": {},
                    "positive_match_data": {"DEFAULT": [{"match_text": "tr:", "start": 0, "stop": 3}]},
                },
            },
        ),
    ],
)
def test_transform_debug_mode(df_emails, expected_result, expected_debug_info):
    """Unit test of the debug mode."""

    # Copy for later load/save test
    df_copy = df_emails.copy()

    # Instanciate manually a detector
    detector = TransferDetector(
        name="transfer",
        header_column="det_clean_header",
        messages_column="messages",
    )

    # Get column names
    res_col = detector.result_column
    debug_dict_col = detector.debug_dict_col

    # Transform data
    df_emails.debug = True
    df_emails = detector.transform(df_emails)

    # Collect results
    result = df_emails[res_col].iloc[0]
    debug_result = df_emails[debug_dict_col].iloc[0]

    # Test result
    assert result == expected_result
    assert debug_result == expected_debug_info


@pytest.mark.parametrize(
    "df, expected_result",
    [
        (
            pd.DataFrame(
                {
                    "from": ["test@gmail.com"],
                    "header": ["tr :Suivi de dossier"],
                    "body": [
                        "",
                    ],
                }
            ),
            True,
        ),
        (
            pd.DataFrame(
                {
                    "from": ["test@gmail.com"],
                    "header": ["fwd: Envoi d'un document de la Société Imaginaire"],
                    "body": [
                        "Bonjour,\n\n\n\n\n\nUn taux d’humidité de 30% a été relevé le 01/01/2024.\n\n\n\nNous reprendrons contact avec l’assurée"
                        + " en Aout 2022.\n\n\n\n\n\n\nBien cordialement,\n\n\n\n\n\nNuméro Auxiliaire : 000000 A / 000000 B\n\n\n\n\n\n\n\n\n"
                        + "SMITH Kim\n\n\nTEST\n-\n\nL\na\nV\nalorisation du\nP\natrimoine\n\n\n2, rue du Test\n\n\n00000 NIORT"
                        + "\n\n\n\n\n\n\n\nTél :    0143740992\n\n\nPort :  0767396737\n\n\nhttp://lvpfrance.fr\n\n\n\n\n\n\n\n\n\n\n\n\n\n\n\n\n"
                        + "\n\n\nDe :\nAccueil - Alex Dupond <accueil@test.fr>\n\n\n\nEnvoyé :\njeudi 01 janvier 2024 01:01\n\n\nÀ "
                        + ":\nCommercial <etudes@test.fr>\n\n\nObjet :\nTR: Evt : A0000000B survenu le 01/01/2024 - Intervention entreprise"
                        + " partenaire\n\n\n\n\n\n\n\n\n\n\n\n\n\n\n\nDe :\n\n\ntest@maif.fr\n[\nmailto:test@maif.fr\n]\n\n\n\n"
                        + "Envoyé :\njeudi 01 janvier 2024 01:01\n\n\nÀ :\nAccueil - Alex Dupond\n\n\nObjet :\nEvt : A0000000B survenu le 01/01/2024"
                        + " - Intervention entreprise partenaire\n\n\n\n\n\nMerci de bien vouloir prendre connaissance du document ci-joint.",
                    ],
                }
            ),
            True,
        ),
        (
            pd.DataFrame(
                {
                    "from": ["test@gmail.com"],
                    "header": ["Virement"],
                    "body": [
                        "Bonjour,\n\n\n\n\n\nUn taux d’humidité de 30% a été relevé le 01/01/2024.\n\n\n\nNous reprendrons contact avec l’assurée"
                        + " en Aout 2022.\n\n\n\n\n\n\nBien cordialement,\n\n\n\n\n\nNuméro Auxiliaire : 000000 A / 000000 B\n\n\n\n\n\n\n\n\n"
                        + "SMITH Kim\n\n\nTEST\n-\n\nL\na\nV\nalorisation du\nP\natrimoine\n\n\n2, rue du Test\n\n\n00000 NIORT"
                        + "\n\n\n\n\n\n\n\nTél :    0143740992\n\n\nPort :  0767396737\n\n\nhttp://lvpfrance.fr\n\n\n\n\n\n\n\n\n\n\n\n\n\n\n\n\n"
                        + "\n\n\nDe :\nAccueil - Alex Dupond <accueil@test.fr>\n\n\n\nEnvoyé :\njeudi 01 janvier 2024 01:01\n\n\nÀ "
                        + ":\nCommercial <etudes@test.fr>\n\n\nObjet :\nTR: Evt : A0000000B survenu le 01/01/2024 - Intervention entreprise"
                        + " partenaire\n\n\n\n\n\n\n\n\n\n\n\n\n\n\n\nDe :\n\n\ntest@maif.fr\n[\nmailto:test@maif.fr\n]\n\n\n\n"
                        + "Envoyé :\njeudi 01 janvier 2024 01:01\n\n\nÀ :\nAccueil - Alex Dupond\n\n\nObjet :\nEvt : A0000000B survenu le 01/01/2024"
                        + " - Intervention entreprise partenaire\n\n\n\n\n\nMerci de bien vouloir prendre connaissance du document ci-joint.",
                    ],
                }
            ),
            False,
        ),
        (
            pd.DataFrame(
                {
                    "from": ["test@gmail.com"],
                    "header": ["tr: virement"],
                    "body": [
                        "Bonjour,\n\n\n\n\n\nUn taux d’humidité de 30% a été relevé le 01/01/2001.\n\n\n\nNous reprendrons contact avec l’assurée"
                        + " en Aout 2022.\n\n\n\n\n\n\nBien cordialement,\n\n\n\n\n\nNuméro : 000000\n\n\n\n\n\n\n\n\nJohn"
                        + " Smith\n\n\nL\na\nValorisation du\nPatrimoine\n\n\n1, rue de la Paix\n\n\n79000 Niort"
                        + "\n\n\n\n\n\n\n\nTél :    0123456789\n\n\nPort :  0123456789\n\n\n\n\n\n\n\n\n\n\n\n\n\n\n"
                        + "\n\n\nDe :\nAccueil - John Smith <john.smith@test.fr>\n\n\n\nEnvoyé :\njeudi 01 janvier 2001 01:01\n\n\nÀ "
                        + ":\nCommercial <test@test.fr>\n\n\nObjet :\nTR: Accident survenu le 01/01/2021 - Intervention"
                        + " partenaire\n\n\n\n\n\n\n\n\n\n\n\n\n\n\n\nDe :\n\n\ntest@test.fr\n[\nmailto:test@test.fr\n]\n\n\n\n"
                        + "Envoyé :\njeudi 01 janvier 2001 01:01\n\n\nÀ :\nAccueil\n\n\nObjet :\nAccident survenu le 01/01/2001"
                        + " - Intervention partenaire\n\n\n\n\n\nMerci de bien vouloir prendre connaissance du document ci-joint.\n\n\n\n"
                        + "Cordialement\n",
                    ],
                }
            ),
            True,
        ),
    ],
)
def test_pipeline_from_config(df, expected_result):
    """
    Instanciate from a config and test the pipeline.
    """
    # Pipeline config key
    pipeline_key = "transfer_pipeline"

    # Create pipeline from config
    pipeline = MelusinePipeline.from_config(config_key=pipeline_key)

    # Apply pipeline on data
    df_transformed = pipeline.transform(df)
    result = df_transformed["transfer_result"][0]

    # Check
    assert result == expected_result



import numpy as np
import pandas as pd
import pytest

from melusine.backend import backend
from melusine.backend.dict_backend import DictBackend
from melusine.backend.pandas_backend import PandasBackend
from melusine.processors import Normalizer


def test_reset_backend():
    """Test"""
    dict_data = {"input_col": "àçöbïù"}
    df_data = pd.DataFrame([dict_data])
    processor = Normalizer(input_columns="input_col", output_columns="output_col")

    dict_backend = DictBackend()
    backend.reset(dict_backend)
    dict_out = processor.transform(dict_data)
    assert isinstance(dict_out, dict)

    backend.reset()
    df_out = processor.transform(df_data)
    assert isinstance(df_out, pd.DataFrame)


def test_unknown_backend():
    with pytest.raises(ValueError):
        backend.reset("unknown")


def test_backend_error():
    backend._backend = None
    with pytest.raises(AttributeError):
        _ = backend.backend
    backend.reset()


def test_check_debug_flag_dict():
    dict_data = {"input_col": "àçöbïù"}
    dict_backend = DictBackend()

    assert dict_backend.check_debug_flag(dict_data) is False

    dict_data = {"debug": False}
    assert dict_backend.check_debug_flag(dict_data) is False

    dict_data = {"debug": True}
    assert dict_backend.check_debug_flag(dict_data) is True


def test_check_debug_flag_pandas():
    df = pd.DataFrame([{"input_col": "àçöbïù"}])
    pandas_backend = PandasBackend()

    assert pandas_backend.check_debug_flag(df) is False

    df.debug = False
    assert pandas_backend.check_debug_flag(df) is False

    df.debug = True
    assert pandas_backend.check_debug_flag(df) is True


def test_add_fields_dict():
    dict_data1 = {"col1": 1, "col8": 8}
    dict_data2 = {"col1": 10, "col2": 2}
    dict_backend = DictBackend()

    data = dict_backend.add_fields(dict_data1, dict_data2)
    assert data == {"col1": 10, "col2": 2, "col8": 8}


def test_dict_backend(backend_base_data, backend_testcase):
    """Test"""
    expected_data = backend_testcase["expected_data"]
    func = backend_testcase["func"]
    input_columns = backend_testcase["input_columns"]
    output_columns = backend_testcase["output_columns"]
    kwargs = backend_testcase.get("kwargs", dict())

    dict_backend = DictBackend()
    n_values = len(list(backend_base_data.values())[0])

    for i in range(n_values):
        data_dict = {key: backend_base_data[key][i] for key in backend_base_data}
        expected_data_dict = {key: expected_data[key][i] for key in expected_data}

        data_dict_transform = dict_backend.apply_transform(
            data=data_dict, func=func, input_columns=input_columns, output_columns=output_columns, **kwargs
        )
        assert data_dict_transform == expected_data_dict


def test_dict_backend_impossible_situation():
    dict_backend = DictBackend()
    with pytest.raises(ValueError):
        _ = dict_backend.apply_transform(
            data={"a": 0}, func=lambda x: x + 1, input_columns=["int_col"], output_columns=None
        )


@pytest.mark.parametrize("progress_bar", [False, True])
def test_pandas_backend(backend_base_data, backend_testcase, progress_bar):
    """Test"""
    expected_data = backend_testcase["expected_data"]
    func = backend_testcase["func"]
    input_columns = backend_testcase["input_columns"]
    output_columns = backend_testcase["output_columns"]
    kwargs = backend_testcase.get("kwargs", dict())

    pandas_backend = PandasBackend(progress_bar=progress_bar, workers=1)

    df_base = pd.DataFrame(backend_base_data)

    df_expected = pd.DataFrame(expected_data)
    df_transform = pandas_backend.apply_transform(
        data=df_base, func=func, input_columns=input_columns, output_columns=output_columns, **kwargs
    )
    pd.testing.assert_frame_equal(df_transform, df_expected)


def test_pandas_backend_multiprocess(backend_base_data, backend_testcase):
    """Test"""
    expected_data = backend_testcase["expected_data"]
    func = backend_testcase["func"]
    input_columns = backend_testcase["input_columns"]
    output_columns = backend_testcase["output_columns"]
    kwargs = backend_testcase.get("kwargs", dict())

    pandas_backend = PandasBackend(progress_bar=False, workers=2)

    # Test on a small dataset (does not trigger multiprocessing)
    df_base = pd.DataFrame(backend_base_data)
    df_expected = pd.DataFrame(expected_data)

    df_transform = pandas_backend.apply_transform(
        data=df_base, func=func, input_columns=input_columns, output_columns=output_columns, **kwargs
    )
    pd.testing.assert_frame_equal(df_transform, df_expected)

    # Augment dataset size to trigger multiprocessing (preserve column type)
    df_long = pd.DataFrame({col: value for col in df_base for value in [np.repeat(df_base[col].values, 3, axis=0)]})
    df_expected_long = pd.DataFrame(
        {col: value for col in df_expected for value in [np.repeat(df_expected[col].values, 3, axis=0)]}
    )

    df_transform = pandas_backend.apply_transform(
        data=df_long, func=func, input_columns=input_columns, output_columns=output_columns, **kwargs
    )
    pd.testing.assert_frame_equal(df_transform, df_expected_long)
